program jcBeamformTD
  ! Program to perform time-domain beamforming
  !
  ! Inputs:
  !   1) history file of traces
  !   2) output slowness and tau parameters
  !   3) history file with xy coordinates
  !
  ! Outputs:
  !   1) history file in t-p domain
  !
  ! Jason Chang
  ! November 2015

  !---------------------------------------
  ! standard declarations
  !---------------------------------------
  use sep
  implicit none


  !---------------------------------------
  ! declare variables
  !---------------------------------------
  ! program parameters
  integer                           :: hh,ii,jj,kk,ll
  integer                           :: itau
  real                              :: xmean,ymean
  real,allocatable,dimension(:)     :: tt,px,py

  ! input parameters
  integer                           :: n1,n2
  real                              :: o1,o2,d1,d2
  real,allocatable,dimension(:)     :: datin
  real,allocatable,dimension(:,:,:) :: datout

  ! t-p parameters
  integer                           :: ntau
  integer                           :: npx,npy
  real                              :: otau,dtau
  real                              :: opx,opy,dpx,dpy
  real                              :: pxmin,pxmax,pymin,pymax

  ! xy parameters
  integer                           :: n1xy,n2xy
  real,allocatable,dimension(:,:)   :: xy


  !---------------------------------------
  ! open I/O
  !---------------------------------------
  call sep_init()

  ! input history file
  call from_history('n1',n1,1 )
  call from_history('n2',n2,1 )
  call from_history('o1',o1,0.)
  call from_history('o2',o2,0.)
  call from_history('d1',d1,1.)
  call from_history('d2',d2,1.)

  ! t-p parameters
  call from_param('ntau',ntau,n1)
  call from_param('otau',otau,o1)
  call from_param('dtau',dtau,d1)
  call from_param("pxmin",pxmin,-1./500.)
  call from_param("pxmax",pxmax, 1./500.)
  call from_param("pymin",pymin,-1./500.)
  call from_param("pymax",pymax, 1./500.)
  call from_param("npx",npx,int(sqrt(real(n2)))  )
  call from_param("dpx",dpx,(pxmax-pxmin)/(npx-1) )
  call from_param("opx",opx,pxmin)
  call from_param("npy",npy,int(sqrt(real(n2)))  )
  call from_param("dpy",dpy,(pymax-pymin)/(npy-1) )
  call from_param("opy",opy,pymin)

  ! xy history file
  if (exist_file('xy_in')) then
    call from_aux('xy_in','n1',n1xy)
    call from_aux('xy_in','n2',n2xy)
    if (n1xy.ne. 2) call seperr('n1xy needs to equal 2')
    if (n2xy.ne.n2) call seperr('n2xy needs to equal n2')
  else
    call seperr('Need xy_in file')
  end if

  ! output history file
  call to_history('n1',ntau)
  call to_history('o1',otau)
  call to_history('d1',dtau)
  call to_history('n2',npx)
  call to_history('o2',opx)
  call to_history('d2',dpx)
  call to_history('n3',npy)
  call to_history('o3',opy)
  call to_history('d3',dpy)


  !---------------------------------------
  ! initializations
  !---------------------------------------
  allocate(datin(n1)); datin=0.
  allocate(datout(ntau,npx,npy)); datout=0.
  allocate(xy(2,n2xy)); xy=0.

  ! read in xy coordinates
  call sreed('xy_in',xy,4*2*n2xy)
  ! center the coordinates
  xmean = sum(xy(1,:))/n2xy
  ymean = sum(xy(2,:))/n2xy
  xy(1,:) = xy(1,:)-xmean
  xy(2,:) = xy(2,:)-ymean

  ! precompute arrays
  allocate(tt(n1))
  do ii = 1,n1
    tt(ii) = o1+(ii-1)*d1
  end do
  allocate(px(npx))
  do ii = 1,npx
    px(ii) = opx+(ii-1)*dpx
  end do
  allocate(py(npy))
  do ii = 1,npy
    py(ii) = opy+(ii-1)*dpy
  end do
  

  !---------------------------------------
  ! kernel
  !---------------------------------------
  do ll = 1,n2
    call sreed('in',datin,4*n1)
    !$OMP PARALLEL DO default(private) shared(xy,px,py,tt,n1,ntau,otau,dtau,datin,datout,ll,npy,npx) schedule(dynamic)
    do kk = 1,npy
      do jj = 1,npx
        do ii = 1,n1
          itau = 1.5 + ((tt(ii)+px(jj)*xy(1,ll)+py(kk)*xy(2,ll)-otau)/dtau)
          if (itau.ge.1 .and. itau.le.ntau) then
            datout(itau,jj,kk) = datout(itau,jj,kk)+datin(ii)
          end if
        end do
      end do
    end do
    !$OMP END PARALLEL DO
  end do

  ! write out
  call srite('out',datout,4*ntau*npx*npy)
        

  !---------------------------------------
  ! clean up
  !---------------------------------------
  call sep_close()
  deallocate(datin,datout,xy,tt,px,py)
  call exit(0)

end program
