program jcNE2RTfull
  ! Program to rotate data from north-east to radial-transverse
  ! Takes correlations between EN and NE into account
  !
  ! Inputs:
  !   1) EE: east correlation history file
  !   2) NN: north correlation history file
  !   3) NE: north source; east receiver
  !   4) EN: east source; north receiver
  !   5) hff: header history file
  !        note - assumed to be the same for both files
  !   6) akey: azimuth key [0]
  !   7) rxkey: receiver x key [0]
  !   8) rykey: receiver y key [0]
  !   9) sxkey: source x key [0]
  !  10) sykey: source y key [0]
  !
  ! Outputs:
  !   1) R: radial correlation history file
  !   2) T: transverse correlation history file
  !
  ! Jason Chang
  ! February 2016


  !======================================
  ! standard declarations
  !======================================
  use sep
  implicit none


  !======================================
  ! declare variables
  !======================================
  ! program variables
  integer                                :: ii
  real                                   :: az,baz,rx,ry,sx,sy
  real                                   :: a1,a2
  real                                   :: b1,b2
  real                                   :: c1,c2,c3,c4
  real,parameter                         :: pi=3.14159265359 

  ! input correlation variables
  integer                                :: ne1,ne2,ne3
  integer                                :: nn1,nn2,nn3
  real                                   :: oe1,oe2,de1,de2
  real                                   :: on1,on2,dn1,dn2
  real,allocatable,dimension(:)          :: ee,nn,en,ne

  ! header variables
  integer                                :: nh1,nh2,nh3
  integer                                :: rxkey,rykey,sxkey,sykey
  real,allocatable,dimension(:)          :: hff

  ! output correlation variables
  real,allocatable,dimension(:)          :: rr,tt



  !======================================
  ! open I/O
  !======================================
  call sep_init()

  ! read in east file
  if (exist_file('EE')) then
    call from_aux('EE','n1',ne1,1 )
    call from_aux('EE','n2',ne2,1 )
    call from_aux('EE','n3',ne3,1 )
    call from_aux('EE','d1',de1,1.)
    call from_aux('EE','d2',de2,1.)
    call from_aux('EE','o1',oe1,0.)
    call from_aux('EE','o2',oe2,0.)
  else
    call seperr('Need east correlation file EE=')
  end if

  ! read in north file
  if (exist_file('NN')) then
    call from_aux('NN','n1',nn1,1 )
    call from_aux('NN','n2',nn2,1 )
    call from_aux('NN','n3',nn3,1 )
    call from_aux('NN','d1',dn1,1.)
    call from_aux('NN','d2',dn2,1.)
    call from_aux('NN','o1',on1,0.)
    call from_aux('NN','o2',on2,0.)
  else
    call seperr('Need north correlation file NN=')
  end if

  ! read in north-east file
  if (exist_file('NE')) then
    call from_aux('NE','n1',nn1,1 )
    call from_aux('NE','n2',nn2,1 )
    call from_aux('NE','n3',nn3,1 )
    call from_aux('NE','d1',dn1,1.)
    call from_aux('NE','d2',dn2,1.)
    call from_aux('NE','o1',on1,0.)
    call from_aux('NE','o2',on2,0.)
  else
    call seperr('Need north-east correlation file NE=')
  end if

  ! read in east-north file
  if (exist_file('EN')) then
    call from_aux('EN','n1',nn1,1 )
    call from_aux('EN','n2',nn2,1 )
    call from_aux('EN','n3',nn3,1 )
    call from_aux('EN','d1',dn1,1.)
    call from_aux('EN','d2',dn2,1.)
    call from_aux('EN','o1',on1,0.)
    call from_aux('EN','o2',on2,0.)
  else
    call seperr('Need east-north correlation file EN=')
  end if

  ! read in header history file
  if (exist_file('hff')) then
    call from_aux('hff','n1',nh1,1 )
    call from_aux('hff','n2',nh2,1 )
    call from_aux('hff','n3',nh3,1 )
  else
    call seperr('Need header history file hff=')
  end if

  ! check???
  if (nn1.ne.ne1) call seperr('n1 needs to match for E and N')
  if (nn2.ne.ne2) call seperr('n2 needs to match for E and N')
  if (nn3.ne.ne3) call seperr('n3 needs to match for E and N')
  if (nh2.ne.ne2) call seperr('n2 needs to match for E and hff')
  if (nh3.ne.ne3) call seperr('n3 needs to match for E and hff')

  ! important header keys
  call from_param('rxkey',rxkey,1)
  call from_param('rykey',rykey,2)
  call from_param('sxkey',sxkey,4)
  call from_param('sykey',sykey,5)

  ! output parameters
  if (exist_file('R')) then
    call to_history('n1',nn1,'R')
    call to_history('n2',nn2,'R')
    call to_history('d1',dn1,'R')
    call to_history('d2',dn2,'R')
    call to_history('o1',on1,'R')
    call to_history('o2',on2,'R')
    call to_history('n3',nn3,'R')
  else
    call seperr('Need radial correlation file R=')
  end if

  if (exist_file('T')) then
    call to_history('n1',ne1,'T')
    call to_history('n2',ne2,'T')
    call to_history('d1',de1,'T')
    call to_history('d2',de2,'T')
    call to_history('o1',oe1,'T')
    call to_history('o2',oe2,'T')
    call to_history('n3',ne3,'T')
  else
    call seperr('Need radial correlation file T=')
  end if


  !======================================
  ! initializations
  !======================================
  allocate(ee(ne1))
  allocate(nn(nn1))
  allocate(ne(ne1))
  allocate(en(nn1))
  allocate(rr(ne1))
  allocate(tt(nn1))
  allocate(hff(nh1))


  !======================================
  ! kernel
  !======================================
  do ii = 1,nn2*nn3
    ! read in files
    call sreed('EE',ee,4*ne1)
    call sreed('NN',nn,4*nn1)
    call sreed('EN',en,4*ne1)
    call sreed('NE',ne,4*nn1)
    call sreed('hff',hff,4*nh1)

    ! calculate rotation coefficients
    rx = hff(rxkey)
    ry = hff(rykey)
    sx = hff(sxkey)
    sy = hff(sykey)
    az = atan2(rx-sx,ry-sy)
    a1 = cos(az)
    a2 = sin(az)
    baz = az+pi
    b1 = cos(baz)
    b2 = sin(baz)
    

    ! apply rotation to data
    c1 = -(a1*b1)
    c2 = a1*b2
    c3 = -(a2*b2)
    c4 = a2*b1
    tt = (c1*ee)+(c2*en)+(c3*nn)+(c4*ne)

    c1 = -(a2*b2)
    c2 = -(a2*b1)
    c3 = -(a1*b1)
    c4 = -(a1*b2)
    rr = (c1*ee)+(c2*en)+(c3*nn)+(c4*ne)

    ! write out
    call srite('R',rr,4*ne1)
    call srite('T',tt,4*nn1)

  end do


  !======================================
  ! clean up
  !======================================
  call sep_close()
  deallocate(ee,nn,rr,tt,hff,en,ne)
  call exit(0)

end program
