program jcPrepNE
  ! Jason P. Chang
  ! March 2013
  ! Update November 2014: now handles odd number of samples
  ! Update March 2016: built for N and E components
  !
  ! similar to original jcBandpass1d, but with 4 axes
  ! fixed the whitening operation
  ! takes N and E as input
  !    whitens by the average smoothed amplitude spectrum
  !
  ! Inputs:
  !   1) Ein: east file
  !   2) Nin: north file
  ! program algorithm:
  !   a) tapers the input trace in time
  !        topt: options for type of taper
  !               0) do nothing
  !               1) rectangle
  !               2) hann-band (default)
  !
  !   b) whitens the spectrum prior to bandpassing, if desired
  !
  !   c) bandpasses a trace in one-dimension
  !        fopt: options for type of filter
  !               0) do nothing
  !               1) rectangle
  !               2) hann-band
  !               3) hann-half (default)
  !        does FFT shifts before FFT and after IFFT
  !        sfftw: single precision (dfftw: double precision)
  !   d) downsample
  !        j1: standard j1 definition
  !
  ! Outputs:
  !   1) Eout: east file
  !   2) Nout: north file
  !   a) result after taper "taper" (optional)
  !   b) filtered spectrum "filt" (optional)


  !========================================
  ! standard declarations
  !========================================
  use sep
  implicit none
  include "fftw3.f"


  !========================================
  ! declare variable types
  !========================================
  ! allocate variables
  integer                           :: ii,jj,kk
  real, parameter                   :: pi=3.1415926535
  real                              :: eps

  ! trace parameters
  integer                           :: nt,n2,n3,n4
  integer                           :: nte,n2e,n3e,n4e
  integer                           :: ntn,n2n,n3n,n4n
  integer                           :: j1,n1out,ntf
  real                              :: ot,dt
  real                              :: ote,dte
  real                              :: otn,dtn

  ! taper parameters
  integer                           :: topt,two_hour
  integer                           :: t1_ind,t2_ind,t3_ind,t4_ind
  real                              :: t1    ,t2    ,t3    ,t4
  real,allocatable,dimension (:)    :: taper

  ! Fourier parameters
  integer                           :: nf
  real                              :: of,df

  ! bandpass parameters
  integer                           :: fopt
  integer                           :: f1_ind,f2_ind,f3_ind,f4_ind
  real                              :: f1    ,f2    ,f3    ,f4
  real,allocatable,dimension (:)    :: filt
  integer*8                         :: fft_for,fft_inv

  real,allocatable,dimension (:)    :: datat,Edatat,Ndatat,Ein,Eout,Nin,Nout,temp
  complex,allocatable, dimension(:) :: dataf,Edataf,Ndataf

  ! whitening parameters
  integer                           :: flen,hflen
  real,allocatable,dimension(:)     :: avg,nampl,Navg,Eampl,Eavg
  logical                           :: white
  

  !========================================
  ! open I/O
  !========================================
  call sep_init()

  !----------------------------------
  ! trace information
  ! east component
  if (exist_file('Ein')) then
    call from_aux('Ein','n1',nte,3600*500*2)
    call from_aux('Ein','d1',dte,.002)
    call from_aux('Ein','o1',ote,0.)
    call from_aux('Ein','n2',n2e,1 )
    call from_aux('Ein','n3',n3e,1 )
    call from_aux('Ein','n4',n4e,1 )
  else
    call seperr('Need Ein')
  end if
  ! north component
  if (exist_file('Nin')) then
    call from_aux('Nin','n1',ntn,3600*500*2)
    call from_aux('Nin','d1',dtn,.002)
    call from_aux('Nin','o1',otn,0.)
    call from_aux('Nin','n2',n2n,1 )
    call from_aux('Nin','n3',n3n,1 )
    call from_aux('Nin','n4',n4n,1 )
  else
    call seperr('Need Nin')
  end if
  ! check parameters are ok
  if (nte.ne.ntn) call seperr('n1 for Ein and Nin do not match')
  if (ote.ne.otn) call seperr('o1 for Ein and Nin do not match')
  if (dte.ne.dtn) call seperr('d1 for Ein and Nin do not match')
  if (n2e.ne.n2n) call seperr('n2 for Ein and Nin do not match')
  if (n3e.ne.n3n) call seperr('n3 for Ein and Nin do not match')
  if (n4e.ne.n4n) call seperr('n4 for Ein and Nin do not match')
  ! pass parameters on
  nt = nte
  ot = ote
  dt = dte
  n2 = n2e
  n3 = n3e
  n4 = n4e

  ! whitening stabilizing parameter
  call from_param('eps',eps,1e-10)

  ! check if odd trace
  if (floor(real(nt)/2) .ne. ceiling(real(nt)/2)) then
    ntf = nt+1
  else
    ntf = nt
  end if

  !----------------------------------
  ! taper parameters
  call from_param('topt',topt,1)
  call from_param('t1'    ,t1    ,ot                )
  call from_param('t1_ind',t1_ind,nint((t1-ot)/dt)+1)
  if (topt.eq.2) then
    call from_param ('t2'    ,t2    ,ot+1*nt/3*dt      )
    call from_param ('t2_ind',t2_ind,nint((t2-ot)/dt)+1)
  else
    call from_param('t2'    ,t2    ,ot+(nt-1)*dt      )
    call from_param('t2_ind',t2_ind,nint((t2-ot)/dt)+1)
  end if
  call from_param ('t3'    ,t3    ,ot+2*nt/3*dt)
  call from_param ('t3_ind',t3_ind,nint((t3-ot)/dt)+1)
  call from_param ('t4'    ,t4    ,ot+(nt-1)*dt)
  call from_param ('t4_ind',t4_ind,nint((t4-ot)/dt)+1)
  ! make sure taper indices make sense
  if (t1_ind.lt.1.) t1_ind= 1
  if (t2_ind.lt.1.) t2_ind= 1
  if (t3_ind.lt.1.) t3_ind= 1
  if (t4_ind.lt.1.) t4_ind= 1
  if (t1_ind.gt.nt) t1_ind=nt
  if (t2_ind.gt.nt) t2_ind=nt
  if (t3_ind.gt.nt) t3_ind=nt
  if (t4_ind.gt.nt) t4_ind=nt

  ! create the taper
  allocate(taper(nt)); taper=0.
  select case(topt)
    case(0)
       ! do nothing
    case(1)
       taper(t1_ind:t2_ind) = 1.
    case(2)
       do ii=t1_ind,t2_ind
         taper(ii) = 0.5*(1.-cos(pi*(ii-t1_ind)/real(t2_ind-t1_ind)))
       end do
       taper(t2_ind:t3_ind)=1.
       do ii=t3_ind,t4_ind
         taper(ii) = 0.5*(1.+cos(pi*(ii-t3_ind)/real(t4_ind-t3_ind)))
       end do
  end select

  if (exist_file('taper')) then
    call to_history('n1',nt,'taper')
    call to_history('d1',dt,'taper')
    call to_history('o1',ot,'taper')
    call to_history('n2',n2,'taper')
    call to_history('n3',n3,'taper')
    call to_history('n4',n4,'taper')
  end if

  !----------------------------------
  ! calculate Fourier units
  nf = ntf/2+1
  of = 0.
  df = 1./(ntf*dt)

  ! filter parameters
  call from_param('fopt',fopt,3)
  call from_param('f1'    ,f1    ,of                )
  call from_param('f1_ind',f1_ind,nint((f1-of)/df)+1)
  if (fopt.eq.2) then
    call from_param('f2'    ,f2    ,of+1*nf/3*df      )
    call from_param('f2_ind',f2_ind,nint((f2-of)/df)+1)
  else
    call from_param('f2'    ,f2    ,of+(nf-1)*df      )
    call from_param('f2_ind',f2_ind,nint((f2-of)/df)+1)
  end if
  call from_param ('f3'    ,f3    ,of+2*nf/3*df      )
  call from_param ('f3_ind',f3_ind,nint((f3-of)/df)+1)
  call from_param ('f4'    ,f4    ,of+(nf-1)*df      )
  call from_param ('f4_ind',f4_ind,nint((f4-of)/df)+1)
  ! make sure filter indices make sense
  if (f1_ind.lt.1.) f1_ind= 1
  if (f2_ind.lt.1.) f2_ind= 1
  if (f3_ind.lt.1.) f3_ind= 1
  if (f4_ind.lt.1.) f4_ind= 1
  if (f1_ind.gt.nf) f1_ind=nf
  if (f2_ind.gt.nf) f2_ind=nf
  if (f3_ind.gt.nf) f3_ind=nf
  if (f4_ind.gt.nf) f4_ind=nf

  ! create the filter
  allocate(filt(nf)); filt=0.

  select case(fopt)
    case(0)
      ! do nothing
    case(1)
      filt(f1_ind:f2_ind) = 1.
    case(2)
      do ii=f1_ind,f2_ind
        filt(ii) = 0.5*(1.-cos(pi*(ii-f1_ind)/real(f2_ind-f1_ind)))
      end do
      filt(f2_ind:f3_ind)=1.
      do ii=f3_ind,f4_ind
        filt(ii) = 0.5*(1.+cos(pi*(ii-f3_ind)/real(f4_ind-f3_ind)))
      end do
    case(3)
      filt(1:f1_ind) = 1.
      do ii=f1_ind,f2_ind
        filt(ii) = 0.5*(1.+cos(pi*(ii-f1_ind)/real(f2_ind-f1_ind)))
      end do
  end select

  if (exist_file('filt')) then
    call to_history('n1', nf ,'filt')
    call to_history('d1', df ,'filt')
    call to_history('o1', of ,'filt')
    call to_history('n2', n2 ,'filt')
    call to_history('n3', n3 ,'filt')
    call to_history('n4', n4 ,'filt')
  end if

  !----------------------------------
  ! whitening paramters
  ! averaging filter total length
  call from_param('white',white,.false.)
  call from_param('flen',flen,21)
  hflen = flen/2

  !----------------------------------
  ! downsample parameter
  call from_param('j1',j1,1)
  n1out = ceiling(real(nt)/j1)

  ! output east
  if (exist_file('Eout')) then
    call to_history('n1',n1out ,'Eout')
    call to_history('n2',n2   ,'Eout')
    call to_history('n3',n3   ,'Eout')
    call to_history('n4',n4   ,'Eout')
    call to_history('d1',j1*dt,'Eout')
    call to_history('o1',ot   ,'Eout')
  else
    call seperr('Need Eout')
  end if
  ! output north
  if (exist_file('Nout')) then
    call to_history('n1',n1out ,'Nout')
    call to_history('n2',n2   ,'Nout')
    call to_history('n3',n3   ,'Nout')
    call to_history('n4',n4   ,'Nout')
    call to_history('d1',j1*dt,'Nout')
    call to_history('o1',ot   ,'Nout')
  else
    call seperr('Need Nout')
  end if


  call sep_close()

  !=========================================
  ! initializations
  !=========================================
  ! allocate
  allocate(Ein   (nt  )); Ein    = 0.
  allocate(Nin   (nt  )); Nin    = 0.
  allocate(datat(ntf )); datat = 0.
  allocate(dataf(nf  )); dataf = 0.
  allocate(Edatat(ntf )); Edatat = 0.
  allocate(Edataf(nf  )); Edataf = 0.
  allocate(Ndatat(ntf )); Ndatat = 0.
  allocate(Ndataf(nf  )); Ndataf = 0.
  allocate(temp (nt  ))
  allocate(Eout  (n1out))
  allocate(Nout  (n1out))
  allocate(Eampl (nf  ))
  allocate(Eavg  (nf  ))
  allocate(Nampl (nf  ))
  allocate(Navg  (nf  ))
  allocate(avg  (nf  ))

  ! set up FFT plans
  call sfftw_plan_dft_r2c_1d(fft_for,nt,datat,dataf,FFTW_ESTIMATE)
  call sfftw_plan_dft_c2r_1d(fft_inv,nt,dataf,datat,FFTW_ESTIMATE)

  ! scale the taper?
  filt = filt/real(ntf)


  !=========================================
  ! kernel
  !=========================================
  do ii=1,n2*n3*n4
    ! read in north and east
    call sreed('Ein',Ein,4*nt)
    call sreed('Nin',Nin,4*nt)


    !----------------------------------
    ! step one: taper
    Ein = Ein*taper
    Nin = Nin*taper

    ! option only to write out E component
    if (exist_file('taper')) call srite('taper',Ein,4*nt)

    !----------------------------------
    ! step two: bandpass
    ! east component
    ! shift into padded data, if necessary
    datat        = 0.
    datat(1:ntf) = Ein
    ! FFT shift the data
    temp(nt-(nt+1)/2+1:nt) = datat(1:(nt+1)/2)
    temp(1:nt-(nt+1)/2)    = datat((nt+1)/2+1:nt)
    datat = temp
    ! FFT the data
    call sfftw_execute(fft_for)
    Edataf = dataf
    !write(0,*) "finished fft"

    ! north component
    ! shift into padded data, if necessary
    datat        = 0.
    datat(1:ntf) = Nin
    ! FFT shift the data
    temp(nt-(nt+1)/2+1:nt) = datat(1:(nt+1)/2)
    temp(1:nt-(nt+1)/2)    = datat((nt+1)/2+1:nt)
    datat = temp
    ! FFT the data
    call sfftw_execute(fft_for)
    Ndataf = dataf

    ! whiten
    if (white) then
      Eavg = 0.
      Eampl = 0.
      Navg = 0.
      Nampl = 0.
      !do jj=1,nf
      !  if ( abs(dataf(jj).ne.0.) ) then
      !    dataf(jj) = dataf(jj)/abs(dataf(jj))
      !  else
      !    dataf(jj) = 0.
      !  end if
      !end do
      do jj=1,nf
        Eampl(jj) = abs(Edataf(jj))
        Nampl(jj) = abs(Ndataf(jj))
      end do

      ! calculate running filter average for each frequency (need to consider BCs)
      ! set the initial
      jj = 1+hflen
      do kk=jj-hflen,jj+hflen
        Eavg(jj) = Eavg(jj) + (Eampl(kk)/real(flen))
        Navg(jj) = Navg(jj) + (Nampl(kk)/real(flen))
      end do

      ! take care of the rest of the middle part
      do jj=2+hflen,nf-hflen
        Eavg(jj) = Eavg(jj-1) - ((Eampl(jj-hflen-1)-Eampl(jj+hflen))/real(flen))
        Navg(jj) = Navg(jj-1) - ((Nampl(jj-hflen-1)-Nampl(jj+hflen))/real(flen))
      end do

      ! now handle boundaries
      do jj=1,hflen
        do kk=1,jj+hflen
          Eavg(jj) = Eavg(jj) + (Eampl(kk)/(real(jj+hflen)+eps))
          Navg(jj) = Navg(jj) + (Nampl(kk)/(real(jj+hflen)+eps))
        end do
      end do
      do jj=nf-hflen+1,nf
        do kk=jj-hflen,nf
          Eavg(jj) = Eavg(jj) + (Eampl(kk)/(real(nf-jj+hflen+1)+eps))
          Navg(jj) = Navg(jj) + (Nampl(kk)/(real(nf-jj+hflen+1)+eps))
        end do
      end do

      ! calculate average smoothed spectrum
      avg = (Eavg+Navg)/2.

      ! normalize by filtered value at given frequency
      do jj=1,nf
        Edataf(jj) = Edataf(jj)/(avg(jj)+eps)
        Ndataf(jj) = Ndataf(jj)/(avg(jj)+eps)
      end do
    end if

    ! filter
    Edataf = Edataf*filt
    Ndataf = Ndataf*filt
    if (exist_file('filt')) call srite('filt',abs(Edataf*sqrt(real(ntf))),4*nf)
    !write(0,*) "finished taper"

    ! IFFT the data
    ! east
    dataf = Edataf
    call sfftw_execute(fft_inv)
    !write(0,*) "finished inverse fft"
    ! IFFT shift the data
    temp(1:(nt+1)/2)    = datat(nt-(nt+1)/2+1:nt)
    temp((nt+1)/2+1:nt) = datat(1:nt-(nt+1)/2)
    datat = temp
    ! swap again
    Ein = datat(1:ntf)
    ! downsample and write out the trace
    Eout = Ein(::j1)
    !write(0,*)"downsampled to",size(out)
    call srite('Eout',Eout,4*n1out)

    ! north
    dataf = Ndataf
    call sfftw_execute(fft_inv)
    !write(0,*) "finished inverse fft"
    ! IFFT shift the data
    temp(1:(nt+1)/2)    = datat(nt-(nt+1)/2+1:nt)
    temp((nt+1)/2+1:nt) = datat(1:nt-(nt+1)/2)
    datat = temp
    ! swap again
    Nin = datat(1:ntf)
    ! downsample and write out the trace
    Nout = Nin(::j1)
    !write(0,*)"downsampled to",size(out)
    call srite('Nout',Nout,4*n1out)
  end do

  ! clean up?
  call sfftw_destroy_plan(fft_for,fft_inv)
  deallocate(datat,dataf,Edatat,Edataf,Ndatat,Ndataf,temp,Ein,Eout,Nin,Nout,Eampl,Nampl,Eavg,Navg,avg)
  call exit(0)

end program
